@@ -207,6 +207,8 @@ validate_and_create_policies(policies_info all_policies, bool if_exists)
 	if (all_policies.refresh && all_policies.refresh->create_policy)
 	{
 		NullableDatum include_tiered_data = { .isnull = true };
+		NullableDatum nbuckets_per_refresh = { .isnull = true };
+		NullableDatum max_batches_per_execution = { .isnull = true };
 
 		if (all_policies.is_alter_policy)
 			policy_refresh_cagg_remove_internal(all_policies.rel_oid, if_exists);
@@ -220,7 +222,9 @@ validate_and_create_policies(policies_info all_policies, bool if_exists)
 														  false,
 														  DT_NOBEGIN,
 														  NULL,
-														  include_tiered_data);
+														  include_tiered_data,
+														  nbuckets_per_refresh,
+														  max_batches_per_execution);
 	}
 	if (all_policies.compress && all_policies.compress->create_policy)
 	{
