@@ -19,9 +19,12 @@
 #include "compat/compat.h"
 
 #include "algorithms/array.h"
+#include "algorithms/bool_compress.h"
 #include "algorithms/deltadelta.h"
 #include "algorithms/dictionary.h"
 #include "algorithms/gorilla.h"
+#include "algorithms/null.h"
+#include "batch_metadata_builder.h"
 #include "chunk.h"
 #include "compression.h"
 #include "create.h"
@@ -31,7 +34,6 @@
 #include "guc.h"
 #include "hypercore/hypercore_handler.h"
 #include "nodes/chunk_dispatch/chunk_insert_state.h"
-#include "segment_meta.h"
 #include "ts_catalog/array_utils.h"
 #include "ts_catalog/catalog.h"
 #include "ts_catalog/compression_chunk_size.h"
@@ -47,6 +49,8 @@ static const CompressionAlgorithmDefinition definitions[_END_COMPRESSION_ALGORIT
 	[COMPRESSION_ALGORITHM_DICTIONARY] = DICTIONARY_ALGORITHM_DEFINITION,
 	[COMPRESSION_ALGORITHM_GORILLA] = GORILLA_ALGORITHM_DEFINITION,
 	[COMPRESSION_ALGORITHM_DELTADELTA] = DELTA_DELTA_ALGORITHM_DEFINITION,
+	[COMPRESSION_ALGORITHM_BOOL] = BOOL_COMPRESS_ALGORITHM_DEFINITION,
+	[COMPRESSION_ALGORITHM_NULL] = NULL_COMPRESS_ALGORITHM_DEFINITION,
 };
 
 static NameData compression_algorithm_name[] = {
@@ -55,6 +59,8 @@ static NameData compression_algorithm_name[] = {
 	[COMPRESSION_ALGORITHM_DICTIONARY] = { "DICTIONARY" },
 	[COMPRESSION_ALGORITHM_GORILLA] = { "GORILLA" },
 	[COMPRESSION_ALGORITHM_DELTADELTA] = { "DELTADELTA" },
+	[COMPRESSION_ALGORITHM_BOOL] = { "BOOL" },
+	[COMPRESSION_ALGORITHM_NULL] = { "NULL" },
 };
 
 Name
@@ -257,7 +263,10 @@ compress_chunk(Oid in_table, Oid out_table, int insert_options)
 	HeapTuple in_table_tp = NULL, index_tp = NULL;
 	Form_pg_attribute in_table_attr_tp, index_attr_tp;
 	CompressionStats cstat;
-	CompressionSettings *settings = ts_compression_settings_get(out_table);
+	/* Might be merging into an existing chunk, so get compression settings
+	 * from that chunk */
+	CompressionSettings *settings = ts_compression_settings_get_by_compress_relid(out_table);
+
 	int64 report_reltuples;
 
 	/* We want to prevent other compressors from compressing this table,
@@ -287,6 +296,7 @@ compress_chunk(Oid in_table, Oid out_table, int insert_options)
 
 	TupleDesc in_desc = RelationGetDescr(in_rel);
 	TupleDesc out_desc = RelationGetDescr(out_rel);
+
 	/* Before calling row compressor relation should be segmented and sorted as configured
 	 * by compress_segmentby and compress_orderby.
 	 * Cost of sorting can be mitigated if we find an existing BTREE index defined for
@@ -603,7 +613,7 @@ compress_chunk_sort_relation(CompressionSettings *settings, Relation in_rel)
 }
 
 void
-compress_chunk_populate_sort_info_for_column(CompressionSettings *settings, Oid table,
+compress_chunk_populate_sort_info_for_column(const CompressionSettings *settings, Oid table,
 											 const char *attname, AttrNumber *att_nums,
 											 Oid *sort_operator, Oid *collation, bool *nulls_first)
 {
@@ -655,7 +665,7 @@ compress_chunk_populate_sort_info_for_column(CompressionSettings *settings, Oid
  * over compressed data
  */
 Oid
-get_compressed_chunk_index(ResultRelInfo *resultRelInfo, CompressionSettings *settings)
+get_compressed_chunk_index(ResultRelInfo *resultRelInfo, const CompressionSettings *settings)
 {
 	int num_segmentby_columns = ts_array_length(settings->fd.segmentby);
 	int num_orderby_columns = ts_array_length(settings->fd.orderby);
@@ -694,7 +704,7 @@ get_compressed_chunk_index(ResultRelInfo *resultRelInfo, CompressionSettings *se
 }
 
 static void
-build_column_map(CompressionSettings *settings, Relation uncompressed_table,
+build_column_map(const CompressionSettings *settings, Relation uncompressed_table,
 				 Relation compressed_table, PerColumn **pcolumns, int16 **pmap)
 {
 	Oid compressed_data_type_oid = ts_custom_type_cache_get(CUSTOM_TYPE_COMPRESSED_DATA)->type_oid;
@@ -742,7 +752,7 @@ build_column_map(CompressionSettings *settings, Relation uncompressed_table,
 			int16 segment_min_attr_offset = segment_min_attr_number - 1;
 			int16 segment_max_attr_offset = segment_max_attr_number - 1;
 
-			SegmentMetaMinMaxBuilder *segment_min_max_builder = NULL;
+			BatchMetadataBuilder *batch_minmax_builder = NULL;
 			if (segment_min_attr_number != InvalidAttrNumber ||
 				segment_max_attr_number != InvalidAttrNumber)
 			{
@@ -750,18 +760,19 @@ build_column_map(CompressionSettings *settings, Relation uncompressed_table,
 					   "could not find the min metadata column");
 				Ensure(segment_max_attr_number != InvalidAttrNumber,
 					   "could not find the min metadata column");
-				segment_min_max_builder =
-					segment_meta_min_max_builder_create(attr->atttypid, attr->attcollation);
+				batch_minmax_builder =
+					batch_metadata_builder_minmax_create(attr->atttypid,
+														 attr->attcollation,
+														 segment_min_attr_offset,
+														 segment_max_attr_offset);
 			}
 
-			Ensure(!is_orderby || segment_min_max_builder != NULL,
+			Ensure(!is_orderby || batch_minmax_builder != NULL,
 				   "orderby columns must have minmax metadata");
 
 			*column = (PerColumn){
 				.compressor = compressor_for_type(attr->atttypid),
-				.min_metadata_attr_offset = segment_min_attr_offset,
-				.max_metadata_attr_offset = segment_max_attr_offset,
-				.min_max_metadata_builder = segment_min_max_builder,
+				.metadata_builder = batch_minmax_builder,
 				.segmentby_column_index = -1,
 			};
 		}
@@ -775,8 +786,6 @@ build_column_map(CompressionSettings *settings, Relation uncompressed_table,
 			*column = (PerColumn){
 				.segment_info = segment_info_new(attr),
 				.segmentby_column_index = index,
-				.min_metadata_attr_offset = -1,
-				.max_metadata_attr_offset = -1,
 			};
 		}
 	}
@@ -788,7 +797,7 @@ build_column_map(CompressionSettings *settings, Relation uncompressed_table,
  ** row_compressor **
  ********************/
 void
-row_compressor_init(CompressionSettings *settings, RowCompressor *row_compressor,
+row_compressor_init(const CompressionSettings *settings, RowCompressor *row_compressor,
 					Relation uncompressed_table, Relation compressed_table,
 					int16 num_columns_in_compressed_table, bool need_bistate, int insert_options)
 {
@@ -885,8 +894,8 @@ row_compressor_process_ordered_slot(RowCompressor *row_compressor, TupleTableSlo
 		row_compressor->first_iteration = false;
 	}
 	bool changed_groups = row_compressor_new_row_is_in_new_group(row_compressor, slot);
-	bool compressed_row_is_full =
-		row_compressor->rows_compressed_into_current_value >= TARGET_COMPRESSED_BATCH_SIZE;
+	bool compressed_row_is_full = row_compressor->rows_compressed_into_current_value >=
+								  (uint32) ts_guc_compression_batch_size_limit;
 	if (compressed_row_is_full || changed_groups)
 	{
 		if (row_compressor->rows_compressed_into_current_value > 0)
@@ -972,21 +981,23 @@ row_compressor_append_row(RowCompressor *row_compressor, TupleTableSlot *row)
 		/* Performance Improvement: Since we call getallatts at the beginning, slot_getattr is
 		 * useless overhead here, and we should just access the array directly.
 		 */
+		BatchMetadataBuilder *builder = row_compressor->per_column[col].metadata_builder;
 		val = slot_getattr(row, AttrOffsetGetAttrNumber(col), &is_null);
 		if (is_null)
 		{
 			compressor->append_null(compressor);
-			if (row_compressor->per_column[col].min_max_metadata_builder != NULL)
-				segment_meta_min_max_builder_update_null(
-					row_compressor->per_column[col].min_max_metadata_builder);
+			if (builder != NULL)
+			{
+				builder->update_null(builder);
+			}
 		}
 		else
 		{
 			compressor->append_val(compressor, val);
-			if (row_compressor->per_column[col].min_max_metadata_builder != NULL)
-				segment_meta_min_max_builder_update_val(row_compressor->per_column[col]
-															.min_max_metadata_builder,
-														val);
+			if (builder != NULL)
+			{
+				builder->update_val(builder, val);
+			}
 		}
 	}
 
@@ -1017,35 +1028,22 @@ row_compressor_flush(RowCompressor *row_compressor, CommandId mycid, bool change
 			Assert(column->segment_info == NULL);
 
 			compressed_data = compressor->finish(compressor);
-
-			/* non-segment columns are NULL iff all the values are NULL */
+			if (compressed_data == NULL)
+			{
+				if (ts_guc_enable_null_compression &&
+					row_compressor->rows_compressed_into_current_value > 0)
+					compressed_data = null_compressor_get_dummy_block();
+			}
 			row_compressor->compressed_is_null[compressed_col] = compressed_data == NULL;
+
 			if (compressed_data != NULL)
 				row_compressor->compressed_values[compressed_col] =
 					PointerGetDatum(compressed_data);
 
-			if (column->min_max_metadata_builder != NULL)
+			if (column->metadata_builder != NULL)
 			{
-				Assert(column->min_metadata_attr_offset >= 0);
-				Assert(column->max_metadata_attr_offset >= 0);
-
-				if (!segment_meta_min_max_builder_empty(column->min_max_metadata_builder))
-				{
-					Assert(compressed_data != NULL);
-					row_compressor->compressed_is_null[column->min_metadata_attr_offset] = false;
-					row_compressor->compressed_is_null[column->max_metadata_attr_offset] = false;
-
-					row_compressor->compressed_values[column->min_metadata_attr_offset] =
-						segment_meta_min_max_builder_min(column->min_max_metadata_builder);
-					row_compressor->compressed_values[column->max_metadata_attr_offset] =
-						segment_meta_min_max_builder_max(column->min_max_metadata_builder);
-				}
-				else
-				{
-					Assert(compressed_data == NULL);
-					row_compressor->compressed_is_null[column->min_metadata_attr_offset] = true;
-					row_compressor->compressed_is_null[column->max_metadata_attr_offset] = true;
-				}
+				column->metadata_builder->insert_to_compressed_row(column->metadata_builder,
+																   row_compressor);
 			}
 		}
 		else if (column->segment_info != NULL)
@@ -1096,20 +1094,9 @@ row_compressor_flush(RowCompressor *row_compressor, CommandId mycid, bool change
 		if (column->compressor != NULL || !column->segment_info->typ_by_val)
 			pfree(DatumGetPointer(row_compressor->compressed_values[compressed_col]));
 
-		if (column->min_max_metadata_builder != NULL)
+		if (column->metadata_builder != NULL)
 		{
-			/* segment_meta_min_max_builder_reset will free the values, so  clear here */
-			if (!row_compressor->compressed_is_null[column->min_metadata_attr_offset])
-			{
-				row_compressor->compressed_values[column->min_metadata_attr_offset] = 0;
-				row_compressor->compressed_is_null[column->min_metadata_attr_offset] = true;
-			}
-			if (!row_compressor->compressed_is_null[column->max_metadata_attr_offset])
-			{
-				row_compressor->compressed_values[column->max_metadata_attr_offset] = 0;
-				row_compressor->compressed_is_null[column->max_metadata_attr_offset] = true;
-			}
-			segment_meta_min_max_builder_reset(column->min_max_metadata_builder);
+			column->metadata_builder->reset(column->metadata_builder, row_compressor);
 		}
 
 		row_compressor->compressed_values[compressed_col] = 0;
@@ -1460,6 +1447,18 @@ decompress_batch(RowDecompressor *decompressor)
 										&decompressor->detoaster,
 										CurrentMemoryContext));
 		CompressedDataHeader *header = get_compressed_data_header(compressed_datum);
+
+		/* Special compression block with the NULL compression algorithm,
+		 * tells that all values in the compressed block are NULLs.
+		 */
+		if (header->compression_algorithm == COMPRESSION_ALGORITHM_NULL)
+		{
+			column_info->iterator = NULL;
+			decompressor->compressed_is_nulls[input_column] = true;
+			decompressor->decompressed_is_nulls[output_index] = true;
+			continue;
+		}
+
 		column_info->iterator =
 			definitions[header->compression_algorithm]
 				.iterator_init_forward(PointerGetDatum(header), column_info->decompressed_type);
@@ -1722,7 +1721,10 @@ tsl_compressed_data_send(PG_FUNCTION_ARGS)
 	pq_begintypsend(&buf);
 	pq_sendbyte(&buf, header->compression_algorithm);
 
-	definitions[header->compression_algorithm].compressed_data_send(header, &buf);
+	if (header->compression_algorithm != COMPRESSION_ALGORITHM_NULL)
+	{
+		definitions[header->compression_algorithm].compressed_data_send(header, &buf);
+	}
 
 	PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
 }
@@ -1738,7 +1740,14 @@ tsl_compressed_data_recv(PG_FUNCTION_ARGS)
 	if (header.compression_algorithm >= _END_COMPRESSION_ALGORITHMS)
 		elog(ERROR, "invalid compression algorithm %d", header.compression_algorithm);
 
-	return definitions[header.compression_algorithm].compressed_data_recv(buf);
+	if (header.compression_algorithm == COMPRESSION_ALGORITHM_NULL)
+	{
+		PG_RETURN_NULL();
+	}
+	else
+	{
+		return definitions[header.compression_algorithm].compressed_data_recv(buf);
+	}
 }
 
 extern Datum
@@ -1830,6 +1839,12 @@ tsl_compressed_data_info(PG_FUNCTION_ARGS)
 		case COMPRESSION_ALGORITHM_ARRAY:
 			has_nulls = array_compressed_has_nulls(header);
 			break;
+		case COMPRESSION_ALGORITHM_BOOL:
+			has_nulls = bool_compressed_has_nulls(header);
+			break;
+		case COMPRESSION_ALGORITHM_NULL:
+			has_nulls = true;
+			break;
 		default:
 			elog(ERROR, "unknown compression algorithm %d", header->compression_algorithm);
 			break;
@@ -1848,6 +1863,40 @@ tsl_compressed_data_info(PG_FUNCTION_ARGS)
 	return HeapTupleGetDatum(tuple);
 }
 
+extern Datum
+tsl_compressed_data_has_nulls(PG_FUNCTION_ARGS)
+{
+	const CompressedDataHeader *header = get_compressed_data_header(PG_GETARG_DATUM(0));
+	bool has_nulls = false;
+
+	switch (header->compression_algorithm)
+	{
+		case COMPRESSION_ALGORITHM_GORILLA:
+			has_nulls = gorilla_compressed_has_nulls(header);
+			break;
+		case COMPRESSION_ALGORITHM_DICTIONARY:
+			has_nulls = dictionary_compressed_has_nulls(header);
+			break;
+		case COMPRESSION_ALGORITHM_DELTADELTA:
+			has_nulls = deltadelta_compressed_has_nulls(header);
+			break;
+		case COMPRESSION_ALGORITHM_ARRAY:
+			has_nulls = array_compressed_has_nulls(header);
+			break;
+		case COMPRESSION_ALGORITHM_BOOL:
+			has_nulls = bool_compressed_has_nulls(header);
+			break;
+		case COMPRESSION_ALGORITHM_NULL:
+			has_nulls = true;
+			break;
+		default:
+			elog(ERROR, "unknown compression algorithm %d", header->compression_algorithm);
+			break;
+	}
+
+	return BoolGetDatum(has_nulls);
+}
+
 extern CompressionStorage
 compression_get_toast_storage(CompressionAlgorithm algorithm)
 {
@@ -1884,6 +1933,12 @@ compression_get_default_algorithm(Oid typeoid)
 		case NUMERICOID:
 			return COMPRESSION_ALGORITHM_ARRAY;
 
+		case BOOLOID:
+			if (ts_guc_enable_bool_compression)
+				return COMPRESSION_ALGORITHM_BOOL;
+			else
+				return COMPRESSION_ALGORITHM_ARRAY;
+
 		default:
 		{
 			/* use dictionary if possible, otherwise use array */
