@@ -4,12 +4,10 @@
  * LICENSE-TIMESCALE for a copy of the license.
  */
 #include <postgres.h>
+
 #include <executor/spi.h>
 #include <fmgr.h>
 #include <lib/stringinfo.h>
-#include <scan_iterator.h>
-#include <scanner.h>
-#include <time_utils.h>
 #include <utils/builtins.h>
 #include <utils/date.h>
 #include <utils/guc.h>
@@ -23,6 +21,9 @@
 #include "debug_assert.h"
 #include "guc.h"
 #include "materialize.h"
+#include "scan_iterator.h"
+#include "scanner.h"
+#include "time_utils.h"
 #include "ts_catalog/continuous_agg.h"
 #include "ts_catalog/continuous_aggs_watermark.h"
 
