@@ -197,8 +197,18 @@ decompress_column(DecompressContext *dcontext, DecompressBatchState *batch_state
 														&dcontext->detoaster,
 														batch_state->per_batch_context));
 
-	/* Decompress the entire batch if it is supported. */
 	CompressedDataHeader *header = (CompressedDataHeader *) value;
+
+	/* First check if this is a block of NULL values. */
+	if (header->compression_algorithm == COMPRESSION_ALGORITHM_NULL)
+	{
+		column_values->decompression_type = DT_Scalar;
+		*column_values->output_isnull = true;
+		*column_values->output_value = (Datum) NULL;
+		return;
+	}
+
+	/* Decompress the entire batch if it is supported. */
 	ArrowArray *arrow = NULL;
 	if (dcontext->enable_bulk_decompression && column_description->bulk_decompression_supported)
 	{
@@ -362,7 +372,9 @@ compressed_batch_get_arrow_array(VectorQualState *vqstate, Expr *expr, bool *is_
 		Assert(column_values->decompression_type != DT_Invalid);
 	}
 
-	Assert(column_values->decompression_type != DT_Iterator);
+	Ensure(column_values->decompression_type != DT_Iterator,
+		   "expected arrow array but got iterator for column index %d",
+		   column_index);
 
 	/*
 	 * Prepare to compute the vector predicate. We have to handle the
