@@ -203,24 +203,6 @@ tsl_set_rel_pathlist_dml(PlannerInfo *root, RelOptInfo *rel, Index rti, RangeTbl
 #endif
 }
 
-/*
- * The fdw needs to expand a distributed hypertable inside the `GetForeignPath`
- * callback. But, since the hypertable base table is not a foreign table, that
- * callback would not normally be called. Thus, we call it manually in this hook.
- */
-void
-tsl_set_rel_pathlist(PlannerInfo *root, RelOptInfo *rel, Index rti, RangeTblEntry *rte)
-{
-	if (is_dummy_rel(rel))
-	{
-		/*
-		 * Don't have to create any other path if the relation is already proven
-		 * to be empty.
-		 */
-		return;
-	}
-}
-
 /*
  * Run preprocess query optimizations
  */
@@ -252,7 +234,7 @@ tsl_postprocess_plan(PlannedStmt *stmt)
 {
 	if (ts_guc_enable_vectorized_aggregation)
 	{
-		stmt->planTree = try_insert_vector_agg_node(stmt->planTree);
+		stmt->planTree = try_insert_vector_agg_node(stmt->planTree, stmt->rtable);
 	}
 
 #ifdef TS_DEBUG
@@ -268,7 +250,9 @@ tsl_postprocess_plan(PlannedStmt *stmt)
 		 */
 		if ((has_normal_agg || has_vector_agg) && (has_vector_agg != should_have_vector_agg))
 		{
-			elog(ERROR, "vector aggregation inconsistent with debug_require_vector_agg GUC");
+			ereport(ERROR,
+					(errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
+					 errmsg("vector aggregation inconsistent with debug_require_vector_agg GUC")));
 		}
 	}
 #endif
